SystemOrganization addCategory: #'Pier-Squeak-Persistency'!

!DateAndTime methodsFor: '*pier-squeak-persistency' stamp: 'lr 9/23/2008 08:13'!
asSortableString
	"A small interpreter to get a date exactly the way required."

	| code |
	code := #( year 4 $- monthIndex 2 $- dayOfMonth 2 $_ hour 2 $- minute 2 $- second 2 ) readStream.
	^ String streamContents: [ :stream |
		[ code atEnd ] whileFalse: [
			stream nextPutAll: ((self perform: code next)
				printPaddedWith: $0 to: code next).
			code next
				ifNotNilDo: [ :separator | stream nextPut: separator ] ] ]! !

PRHistoryPersistency subclass: #PRDelayedPersistency
	instanceVariableNames: 'duration process message'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Squeak-Persistency'!

!PRDelayedPersistency class methodsFor: 'descriptions' stamp: 'lr 1/29/2009 21:32'!
descriptionDuration
	^ MADurationDescription new
		comment: 'The minimum interval between two snapshots.';
		accessor: #duration;
		label: 'Duration';
		default: 1 hour;
		priority: 100;
		beRequired;
		yourself! !

!PRDelayedPersistency class methodsFor: 'descriptions' stamp: 'lr 1/29/2009 21:18'!
descriptionMessage
	^ MAStringDescription new
		accessor: #message;
		label: 'Message';
		group: 'Status';
		priority: 200;
		beReadonly;
		yourself! !

!PRDelayedPersistency class methodsFor: 'testing' stamp: 'lr 1/29/2009 21:15'!
isAbstract
	^ self name = #PRDelayedPersistency! !

!PRDelayedPersistency methodsFor: 'accessing' stamp: 'lr 1/29/2009 21:16'!
duration
	^ duration! !

!PRDelayedPersistency methodsFor: 'accessing' stamp: 'lr 1/29/2009 21:16'!
duration: aDuration
	"The minimum interval between two snapshots."

	duration := aDuration! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 1/29/2009 21:20'!
execute: aCommand
	self mutex critical: [
		self apply: aCommand.
		aCommand isLogged ifTrue: [
			self add: aCommand context.
			self log: aCommand context ] ].
	self start! !

!PRDelayedPersistency methodsFor: 'testing' stamp: 'lr 1/29/2009 21:26'!
isScheduled
	^ process notNil and: [ process isTerminated not ]! !

!PRDelayedPersistency methodsFor: 'private' stamp: 'lr 1/29/2009 21:21'!
log: aContext
	| stream |
	stream := FileDirectory default
		fileNamed: 'transactions.txt'.
	[ self log: aContext on: stream setToEnd ]
		ensure: [ stream close ]! !

!PRDelayedPersistency methodsFor: 'private' stamp: 'lr 1/29/2009 21:21'!
log: aContext on: aStream
	aStream nextPutAll: ' == '; nextPutAll: aContext command class label; nextPutAll: ' == '; cr.
	aStream nextPutAll: ' -- context -- '; cr.
	aContext description do: [ :each |
		aStream
			nextPutAll: each label; nextPutAll: ': ';
			nextPutAll: (each toString: (aContext
				readUsing: each)); cr ].
	aStream nextPutAll: ' -- command -- '; cr.
	aContext command description do: [ :each |
		aStream
			nextPutAll: each label; nextPutAll: ': ';
			nextPutAll: (each toString: (aContext command
				readUsing: each)); cr ].
	aStream cr; cr! !

!PRDelayedPersistency methodsFor: 'accessing' stamp: 'lr 2/1/2009 16:52'!
message
	^ message! !

!PRDelayedPersistency methodsFor: 'accessing' stamp: 'lr 2/1/2009 16:52'!
message: aString
	message := aString! !

!PRDelayedPersistency methodsFor: 'private' stamp: 'lr 1/29/2009 22:01'!
process
	(Delay forDuration: self duration)
		wait.
	Processor activeProcess
		priority: Processor highestPriority.
	process := nil.
	self snapshot! !

!PRDelayedPersistency methodsFor: 'initialization' stamp: 'lr 1/29/2009 21:20'!
setKernel: aKernel
	super setKernel: aKernel.
	self stop! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 1/29/2009 21:31'!
snapshot
	"Take a snapshot and perform cleanup tasks."

	self subclassResponsibility! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 1/29/2009 21:30'!
start
	"Schedule the receiver to take a snapshot after the configured duration."

	self isScheduled
		ifTrue: [ ^ self ].
	process := [ self process ]
		newProcess.
	process 
		priority: Processor userBackgroundPriority;
		name: 'pier'.
	process resume! !

!PRDelayedPersistency methodsFor: 'actions' stamp: 'lr 1/29/2009 21:31'!
stop
	"Stop the configured snapshot."

	self isScheduled
		ifTrue: [ process terminate ].
	process := nil! !

PRDelayedPersistency subclass: #PRImagePersistency
	instanceVariableNames: 'namePrefix maxImages'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Squeak-Persistency'!

!PRImagePersistency commentStamp: 'lr 6/22/2008 17:53' prior: 0!
I am a persistency strategy regularly saving the whole image. Do not use me when developing, as this might corrupt the changes file.!

!PRImagePersistency class methodsFor: 'accessing-descriptions' stamp: 'lr 1/23/2009 14:02'!
descriptionMaxImages
	^ MANumberDescription new
		comment: 'The maximum number of backup copies to keep';
		accessor: #maxImages;
		label: 'Backup Images';
		priority: 110;
		beRequired;
		default: 3;
		beInteger;
		yourself! !

!PRImagePersistency class methodsFor: 'accessing-descriptions' stamp: 'lr 1/23/2009 14:03'!
descriptionNamePrefix
	^ MAStringDescription new
		comment: 'Defines the prefix for image-names.';
		default: 'pier_backup_';
		accessor: #namePrefix;
		label: 'Backup Prefix';
		priority: 120;
		beRequired;
		yourself! !

!PRImagePersistency methodsFor: 'private' stamp: 'lr 1/29/2009 21:31'!
cleanupBackupImages
	| images |
	maxImages isNil ifTrue: [ ^ self ].
	images := FileDirectory default fileNames select: [ :each | 
		(each endsWith: FileDirectory imageSuffix) 
			and: [ each beginsWith: namePrefix ] ].
	images size <= maxImages ifTrue: [ ^ self ].
	images := (images asSortedCollection allButLast: maxImages)
		collect: [ :each | each copyUpToLast: $. ].
	images do: [ :each |
		FileDirectory default 
			deleteFileNamed: (each , '.' , FileDirectory imageSuffix)
			ifAbsent: [ ].
		FileDirectory default 
			deleteFileNamed: (each , '.' , FileDirectory changeSuffix)
			ifAbsent: [ ] ]! !

!PRImagePersistency methodsFor: 'accessing' stamp: 'lr 6/22/2008 17:54'!
maxImages
	^ maxImages! !

!PRImagePersistency methodsFor: 'accessing' stamp: 'lr 1/29/2009 21:27'!
maxImages: anInteger
	"The maximum number of backup copies to keep."

	maxImages := anInteger! !

!PRImagePersistency methodsFor: 'accessing' stamp: 'lr 6/22/2008 18:06'!
namePrefix
	^ namePrefix! !

!PRImagePersistency methodsFor: 'accessing' stamp: 'lr 1/29/2009 21:27'!
namePrefix: aString
	"Defines the prefix for backup copies."

	namePrefix := aString! !

!PRImagePersistency methodsFor: 'private' stamp: 'lr 1/23/2009 13:58'!
saveImageAndBackupAs: aString
	| image |
	image := SmalltalkImage current.
	image closeSourceFiles.
	FileDirectory default
		copyFileWithoutOverwriteConfirmationNamed: SmalltalkImage current imageName 
		toFileNamed: (FileDirectory default pathName , FileDirectory slash , aString , '.' , FileDirectory imageSuffix).
	FileDirectory default
		copyFileWithoutOverwriteConfirmationNamed: SmalltalkImage current changesName 
		toFileNamed: (FileDirectory default pathName , FileDirectory slash , aString , '.' , FileDirectory changeSuffix).	
	Smalltalk setMacFileInfoOn: (FileDirectory default pathName , FileDirectory slash , aString , '.' , FileDirectory changeSuffix).
	image openSourceFiles; saveImageSegments; snapshot: true andQuit: false! !

!PRImagePersistency methodsFor: 'actions' stamp: 'lr 1/29/2009 21:35'!
snapshot
	| filename |
	filename := namePrefix , DateAndTime now asSortableString.
	self saveImageAndBackupAs: filename.
	self message: 'Last snapshot ' , filename printString , ' at ' , TimeStamp now asString.
	self cleanupBackupImages! !

PRDelayedPersistency subclass: #PRSegmentPersistency
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Squeak-Persistency'!

!PRSegmentPersistency class methodsFor: 'testing' stamp: 'lr 1/30/2009 16:55'!
isAbstract
	^ true! !

!PRSegmentPersistency methodsFor: 'private' stamp: 'lr 1/30/2009 08:14'!
cleanup
	"Execute all the registered cleanup code. Do not execute this code within an image you intend to continue using. This is a non-reversible operation and pretty much destroys your complete environment."

	(Pragma allNamed: #cleanup in: self class)
		do: [ :pragma | self perform: pragma selector ].
	Smalltalk garbageCollect! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:04'!
cleanupKom
	"Stop all the server adaptors, they just get into our way."
	<cleanup>
	
	WAKom withAllSubclassesDo: [ :each | each shutDown ]! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:03'!
cleanupMagritte
	"Clear the magritte cache that shares common description objects with the kernel."
	<cleanup>
	
	MADescriptionBuilder initialize! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:21'!
cleanupMonticello
	"Clear monticello caches, they might reference some of our blocks in obscure ways."
	<cleanup>
	
	MCMethodDefinition shutDown.
	MCFileBasedRepository flushAllCaches! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:04'!
cleanupPier
	"Clear all registered kernels. We don't want references to our model from outside."
	<cleanup>
	
	PRKernel instVarNamed: 'Instances' put: nil! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:13'!
cleanupProcesses
	"Get rid of pocesses that potentialy still hang somewhere and reference our model. Take care not to kill any of the vital processes."
	<cleanup>

	| protected |
	protected := IdentitySet new.
	protected
		add: Project uiProcess;
		add: Processor activeProcess;
		add: Sensor eventTicklerProcess;
		add: Processor backgroundProcess;
		add: Sensor interruptWatcherProcess;
		add: Smalltalk lowSpaceWatcherProcess;
		add: WeakArray runningFinalizationProcess.
	Process allInstances do: [ :each |
		(each isTerminated or: [ Processor highestPriority = each priority or: [ protected includes: each ] ])
			ifFalse: [ each terminate ] ]
	.
	! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:10'!
cleanupSeaside
	"Clear all the entry points that reference other entry points, state and continuations by overriding them with a blank version. Do not try to call #clearAllHandlers because this spawns a new process. Also dump the default dispatcher just to be sure."
	<cleanup>

	WAEntryPoint allSubInstances
		do: [ :entry | entry copyFrom: entry class basicNew ].
	WADispatcher instVarNamed: 'default' put: nil! !

!PRSegmentPersistency methodsFor: 'cleanup' stamp: 'lr 1/30/2009 08:18'!
cleanupWorld
	"Close all windows, just in case we have some workspaces or pier browsers open that reference our model."
	<cleanup>

	World submorphs do: [ :morph | morph delete ].
	World releaseCachedState! !

!PRSegmentPersistency methodsFor: 'private' stamp: 'lr 1/30/2009 08:34'!
serialize: aString
	| stream pointers |
	stream := self directory forceNewFileNamed: aString.
	pointers := [ [ ImageSegment new writeKernel: self kernel on: stream ]
		ensure: [ stream close ] ] on: Exception do: [ :error | nil ].
	(pointers isNil or: [ pointers anySatisfy: [ :each | (each isLiteral or: [ each isVariableBinding ]) not ] ]) 
		ifTrue: [ self directory deleteFileNamed: aString ifAbsent: [ ] ]! !

!PRSegmentPersistency methodsFor: 'actions' stamp: 'lr 1/29/2009 22:02'!
snapshot
	| filename |
	OSProcess isNil
		ifTrue: [ ^ self message: 'OSProcess not installed.' ].
	OSProcess thisOSProcess pid
		ifTrue: [ ^ self message: 'OSProcess does not support this platform.' ].
	filename := name , DateAndTime now asSortableString , '.obj'.
	OSProcess thisOSProcess forkHeadlessSqueakAndDo: [
		[ self cleanup; serialize: filename ]
			ensure: [ Smalltalk quitPrimitive ] ].
	self message: 'Last snapshot ' , filename printString , ' at ' , TimeStamp now asString! !

!ImageSegment methodsFor: '*Pier-Squeak-Persistency' stamp: 'lr 1/30/2009 08:26'!
writeKernel: aKernel on: aStream
	"Serialize aKernel on aStream so that it can be loaded into another image later. Answer a list of outpointers, these are objects that are referenced from somewhere else."

	| previousMarker |
	self
		copyFromRoots: (Array with: aKernel)
		sizeHint: 1024 *1024 * 1024
		areUnique: true.
	state = #activeCopy
		ifFalse: [ self error: 'The image segment has an invalid state.' ].
	previousMarker := endMarker.
	endMarker := nil.
	aStream fileOutClass: nil andObject: self.
	endMarker := previousMarker.
	^ outPointers copy! !

PRPersistencyTest subclass: #PRImagePersistencyTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Pier-Squeak-Persistency'!

!PRImagePersistencyTest class methodsFor: 'testing' stamp: 'lr 3/30/2007 22:22'!
isAbstract
	^ false! !

!PRImagePersistencyTest methodsFor: 'private' stamp: 'lr 3/30/2007 22:22'!
actualClass
	^ PRImagePersistency! !
